/* This file implements all calls made to dos-specific code using SDL */
#include <stdlib.h>
#include <assert.h>
#include <SDL/SDL.h>
#ifdef DOSEMU_CAPTURE
#include <imago2.h>
#endif
#include "gfx.h"
#include "mouse.h"
#include "keyb.h"
#include "timer.h"

static void cleanup(void);
static void proc_events(void);
static int translate_sdlkey(int sym);
static void update_modkeys(void);
static void capture_frame(unsigned char *frame);

static void init_sdl(void)
{
	const SDL_version *ver;

	if(!SDL_WasInit(SDL_INIT_EVERYTHING)) {
		SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);

		if((ver = SDL_Linked_Version())) {
			printf("SDL %d.%d.%d initialized\n", ver->major, ver->minor, ver->patch);
		}
		atexit(cleanup);
	}
}

static void cleanup(void)
{
	SDL_Quit();
}

/* ----- graphics (gfx.c (vbe) implementation) ----- */
static SDL_Surface *fbsurf;
static int scale = 1;
static int frames_to_capture;

static struct {
	unsigned char r, g, b;
} palette[256];


void *set_video_mode(int xsz, int ysz, int bpp)
{
	unsigned int sdl_flags = bpp <= 8 ? SDL_HWPALETTE : 0;
	char *env;

	if(getenv("DOSEMU_DOUBLESIZE")) {
		scale = 2;
	}

	if((env = getenv("DOSEMU_SCALE"))) {
		int n = atoi(env);
		if(n > 0) {
			scale = n;
		}
	}
	xsz *= scale;
	ysz *= scale;

	if(getenv("DOSEMU_FULLSCREEN")) {
		sdl_flags |= SDL_FULLSCREEN;
	}

	init_sdl();

	if(!(fbsurf = SDL_SetVideoMode(xsz, ysz, bpp, sdl_flags))) {
		fprintf(stderr, "failed to set video mode\n");
		abort();
	}
	SDL_WM_SetCaption("Rayzor", 0);
	/*SDL_ShowCursor(0);*/

	return fbsurf->pixels;
}

int set_text_mode(void)
{
	SDL_ShowCursor(1);
	SDL_EnableKeyRepeat(0, 0);
	return 0;
}

int get_color_depth(void)
{
	return fbsurf->format->BitsPerPixel;
}

static int count_bits(unsigned int x)
{
	int i, count = 0;
	for(i=0; i<32; i++) {
		if(x & 1) {
			count++;
		}
		x >>= 1;
	}
	return count;
}

int get_color_bits(int *rbits, int *gbits, int *bbits)
{
	*rbits = count_bits(fbsurf->format->Rmask);
	*gbits = count_bits(fbsurf->format->Gmask);
	*bbits = count_bits(fbsurf->format->Bmask);
	return *rbits + *gbits + *bbits;
}

int get_color_shift(int *rshift, int *gshift, int *bshift)
{
	*rshift = fbsurf->format->Rshift;
	*gshift = fbsurf->format->Gshift;
	*bshift = fbsurf->format->Bshift;
	return 0;
}

int get_color_mask(unsigned int *rmask, unsigned int *gmask, unsigned int *bmask)
{
	*rmask = fbsurf->format->Rmask;
	*gmask = fbsurf->format->Gmask;
	*bmask = fbsurf->format->Bmask;
	return 0;
}


void set_palette(int idx, int r, int g, int b)
{
	SDL_Color col;
	col.r = r;
	col.g = g;
	col.b = b;

	if(SDL_SetPalette(fbsurf, SDL_LOGPAL | SDL_PHYSPAL, &col, idx, 1) != 1) {
		fprintf(stderr, "set_palette failed to set the required color\n");
	}

	palette[idx].r = r;
	palette[idx].g = g;
	palette[idx].b = b;
}

static int copy_frame_called;

void copy_frame(void *pixels)
{
	unsigned char *frame = (unsigned char*)pixels;
	copy_frame_called = 1;

	if(SDL_MUSTLOCK(fbsurf)) {
		SDL_LockSurface(fbsurf);
	}

	if(pixels) {
		if(scale > 1) {
			int i, j, xsz, ysz;
			unsigned char *dest = fbsurf->pixels;

			xsz = fbsurf->w * scale;
			ysz = fbsurf->h * scale;

			for(i=0; i<ysz; i++) {
				for(j=0; j<xsz; j++) {
					*dest++ = frame[(i / scale) * fbsurf->w + (j / scale)];
				}
			}
		} else {
			int num_pixels = fbsurf->w * fbsurf->h;
			memcpy(fbsurf->pixels, frame, num_pixels * fbsurf->format->BytesPerPixel);
		}
	}

	if(SDL_MUSTLOCK(fbsurf)) {
		SDL_UnlockSurface(fbsurf);
	}

	SDL_Flip(fbsurf);

	if(frames_to_capture > 0) {
		capture_frame(pixels);
		--frames_to_capture;
	}

	/* also print fps every second ... */
	/*{
		static long prev_fps, num_frames;
		long msec, dt;

		msec = get_msec();
		dt = msec - prev_fps;
		if(dt >= 1000) {
			float fps = (float)num_frames / ((float)dt / 1000.0f);
			printf("framerate: %.1f      \r", fps);
			fflush(stdout);
			num_frames = 0;
			prev_fps = msec;
		} else {
			num_frames++;
		}
	}*/
}

#define spin_delay(ms) \
	do { \
		unsigned int end = SDL_GetTicks() + ms; \
		while((prev_msec = SDL_GetTicks()) < end); \
	} while(0)

#define FRAME_INTERVAL	(1000/70)

void wait_vsync(void)
{
	/*static int prev_msec;
	int msec, dt, tleft;*/

	if(!copy_frame_called) {
		copy_frame(0);
		copy_frame_called = 0;
	}

	/*
	msec = SDL_GetTicks();

	dt = msec - prev_msec;

	tleft = FRAME_INTERVAL - dt;
	if(tleft > 0) {
		int coarse = tleft & 0xfffffff8;
		tleft = tleft & 7;

		if(coarse) {
			SDL_Delay(coarse);
		}
		if(tleft) {
			spin_delay(tleft);
		} else {
			prev_msec = SDL_GetTicks();
		}
	} else {
		prev_msec = msec;
	}
	*/
}

static int cap_count = 0;
void begin_capture(int frames)
{
	frames_to_capture = frames;
}

void end_capture(void)
{
	cap_count = 0;
	frames_to_capture = 0;
}

#define NUMPIX	(fbsurf->w * fbsurf->h)
static void capture_frame(unsigned char *frame)
{
#ifdef DOSEMU_CAPTURE
	static unsigned char rgbpix[NUMPIX * 4];
	char fname[32];
	int i;
	unsigned char *src, *dest;

	sprintf(fname, "frame%04d.png", cap_count++);

	src = frame;
	dest = rgbpix;

	for(i=0; i<NUMPIX; i++) {
		unsigned char c = *src++;
		*dest++ = palette[c].r;
		*dest++ = palette[c].g;
		*dest++ = palette[c].b;
		*dest++ = 255;
	}

	img_save_pixels(fname, rgbpix, fbsurf->w, fbsurf->h, IMG_FMT_RGBA32);
#endif
}


/* ----- event handling (conio.h) ----- */
static SDL_Event *keybev;
static int mousex, mousey, bnmask;

static int keystate[256];
static int num_pressed;

int kbhit(void)
{
	if(!keybev) {
		proc_events();
	}
	return keybev != 0;
}

int getch(void)
{
	int res;

	while(!keybev) {
		SDL_Event ev;
		SDL_WaitEvent(&ev);
		SDL_PushEvent(&ev);
		proc_events();
	}
	res = keybev->key.keysym.sym;
	keybev = 0;
	return res;
}

/* ----- improved event handling (keyb.h) ---- */
static char sdlkey_tbl[SDLK_LAST];

int kb_init(int bufsz)
{
	int i;
	init_sdl();

	/* construct the SDL key translation table */
	for(i=0; i<256; i++) {
		sdlkey_tbl[i] = i;	/* start from an identity mapping for the first 256 */
	}
	/* then change the special keys */
	sdlkey_tbl[SDLK_LALT] = KB_LALT;
	sdlkey_tbl[SDLK_RALT] = KB_RALT;
	sdlkey_tbl[SDLK_LCTRL] = KB_LCTRL;
	sdlkey_tbl[SDLK_RCTRL] = KB_RCTRL;
	sdlkey_tbl[SDLK_LSHIFT] = KB_LSHIFT;
	for(i=0; i<12; i++) {
		sdlkey_tbl[SDLK_F1 + i] = KB_F1 + i;
	}
	sdlkey_tbl[SDLK_CAPSLOCK] = KB_CAPSLK;
	sdlkey_tbl[SDLK_NUMLOCK] = KB_NUMLK;
	sdlkey_tbl[SDLK_SCROLLOCK] = KB_SCRLK;
	sdlkey_tbl[SDLK_SYSREQ] = KB_SYSRQ;
	sdlkey_tbl[SDLK_ESCAPE] = KB_ESC;
	sdlkey_tbl[SDLK_INSERT] = KB_INSERT;
	sdlkey_tbl[SDLK_DELETE] = KB_DEL;
	sdlkey_tbl[SDLK_HOME] = KB_HOME;
	sdlkey_tbl[SDLK_END] = KB_END;
	sdlkey_tbl[SDLK_PAGEUP] = KB_PGUP;
	sdlkey_tbl[SDLK_PAGEDOWN] = KB_PGDN;
	sdlkey_tbl[SDLK_LEFT] = KB_LEFT;
	sdlkey_tbl[SDLK_RIGHT] = KB_RIGHT;
	sdlkey_tbl[SDLK_UP] = KB_UP;
	sdlkey_tbl[SDLK_DOWN] = KB_DOWN;
	sdlkey_tbl[SDLK_KP_PERIOD] = KB_NUM_DOT;
	sdlkey_tbl[SDLK_KP_ENTER] = KB_NUM_ENTER;
	sdlkey_tbl[SDLK_KP_PLUS] = KB_NUM_PLUS;
	sdlkey_tbl[SDLK_KP_MINUS] = KB_NUM_MINUS;
	sdlkey_tbl[SDLK_KP_MULTIPLY] = KB_NUM_MUL;
	sdlkey_tbl[SDLK_KP_DIVIDE] = KB_NUM_DIV;
	/* TODO missing numeric keypad numbers */
	sdlkey_tbl[SDLK_BACKSPACE] = KB_BACKSP;


	return 0;
}

void kb_shutdown(void)
{
}

int kb_getkey(void)
{
	int res = -1;

	proc_events();
	if(keybev) {
		res = translate_sdlkey(keybev->key.keysym.sym);
		keybev = 0;
	}
	return res;
}

int kb_isdown(int key)
{
	switch(key) {
	case KB_ANY:
		return num_pressed;

	case KB_ALT:
		return keystate[KB_LALT] + keystate[KB_RALT];

	case KB_CTRL:
		return keystate[KB_LCTRL] + keystate[KB_RCTRL];
	}
	return keystate[key];
}

/* mouse handling (mouse.c implementation) */
static unsigned long last_mouse_hide_time;

int have_mouse(void)
{
	return 1;
}

void set_mouse_limits(int xmin, int ymin, int xmax, int ymax)
{
}

void set_mouse(int x, int y)
{
	SDL_ShowCursor(0);
	last_mouse_hide_time = get_msec();

	SDL_WarpMouse(x * scale, y * scale);
	mousex = x;
	mousey = y;
}

int read_mouse(int *xp, int *yp)
{
	if(xp) *xp = mousex;
	if(yp) *yp = mousey;
	return bnmask;
}

static void proc_events(void)
{
	static SDL_Event ev;

	if(last_mouse_hide_time > 0 && get_msec() - last_mouse_hide_time > 3000) {
		last_mouse_hide_time = 0;
		SDL_ShowCursor(1);
	}

	while(SDL_PollEvent(&ev)) {
		switch(ev.type) {
		case SDL_KEYDOWN:
			{
				int key = translate_sdlkey(ev.key.keysym.sym);

				if(!keybev) {
					keybev = &ev;
				}

				if(!keystate[key]) {
					keystate[key] = 1;
					num_pressed++;
				}

				update_modkeys();
			}
			break;

		case SDL_KEYUP:
			{
				int key = translate_sdlkey(ev.key.keysym.sym);

				if(keystate[key]) {
					keystate[key] = 0;
					if(--num_pressed < 0) {
						num_pressed = 0;
					}
				}

				update_modkeys();
			}
			break;

		case SDL_MOUSEMOTION:
			mousex = ev.motion.x / scale;
			mousey = ev.motion.y / scale;
			break;

		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP:
			{
				int mask = 0;
				switch(ev.button.button) {
				case SDL_BUTTON_LEFT:
					mask = MOUSE_LEFT;
					break;
				case SDL_BUTTON_MIDDLE:
					mask = MOUSE_MIDDLE;
					break;
				case SDL_BUTTON_RIGHT:
					mask = MOUSE_RIGHT;
				default:
					break;
				}
				if(!mask) {
					break;
				}

				if(ev.button.state == SDL_PRESSED) {
					bnmask |= mask;
				} else {
					bnmask &= ~mask;
				}
			}
			break;

		default:
			break;
		}
	}
}

static int translate_sdlkey(int sym)
{
	return sdlkey_tbl[sym];
}

static void update_modkeys(void)
{
	unsigned int mod = SDL_GetModState();

	keystate[KB_LALT] = mod & KMOD_LALT;
	keystate[KB_RALT] = mod & KMOD_RALT;
	keystate[KB_LCTRL] = mod & KMOD_LCTRL;
	keystate[KB_RCTRL] = mod & KMOD_RCTRL;
	keystate[KB_LSHIFT] = mod & KMOD_LSHIFT;
	keystate[KB_RSHIFT] = mod & KMOD_RSHIFT;
}

/* ---- timer.c implementation ---- */
static Uint32 start_time;

void init_timer(int res_hz)
{
	init_sdl();
	reset_timer();
}

void reset_timer(void)
{
	start_time = SDL_GetTicks();
	printf("resetting timer: %u, %lu\n", start_time, get_msec());
}

unsigned long get_msec(void)
{
	Uint32 ticks = SDL_GetTicks();
	return (unsigned long)(ticks - start_time);
}
