#include "lines.h"

enum {
	IN		= 0,
	LEFT	= 1,
	RIGHT	= 2,
	TOP		= 4,
	BOTTOM	= 8
};

static int outcode(int x, int y, int minx, int miny, int maxx, int maxy)
{
	int code = 0;

	if(x < minx) {
		code |= LEFT;
	} else if(x > maxx) {
		code |= RIGHT;
	}
	if(y < miny) {
		code |= TOP;
	} else if(y > maxy) {
		code |= BOTTOM;
	}
	return code;
}

#define FIXMUL(a, b)	(((a) * (b)) >> 8)
#define FIXDIV(a, b)	(((a) << 8) / (b))

#define LERP(a, b, t)	((a) + FIXMUL((b) - (a), (t)))

int clip_line2d(int *x0, int *y0, int *x1, int *y1, const int *vp)
{
	int oc_out;
	int minx = vp[0];
	int miny = vp[1];
	int maxx = vp[0] + vp[2] - 1;
	int maxy = vp[1] + vp[3] - 1;

	int oc0 = outcode(*x0, *y0, minx, miny, maxx, maxy);
	int oc1 = outcode(*x1, *y1, minx, miny, maxx, maxy);

	int32_t fx0, fy0, fx1, fy1, fminx, fminy, fmaxx, fmaxy;

	if(!(oc0 | oc1)) return 1;	/* both points are inside */

	fx0 = *x0 << 8;
	fy0 = *y0 << 8;
	fx1 = *x1 << 8;
	fy1 = *y1 << 8;
	fminx = minx << 8;
	fminy = miny << 8;
	fmaxx = maxx << 8;
	fmaxy = maxy << 8;

	for(;;) {
		int32_t x, y, t;

		if(oc0 & oc1) return 0;		/* both have points with the same outbit, not visible */
		if(!(oc0 | oc1)) break;		/* both points are inside */

		oc_out = oc0 ? oc0 : oc1;

		if(oc_out & TOP) {
			t = FIXDIV(fminy - fy0, fy1 - fy0);
			x = LERP(fx0, fx1, t);
			y = fminy;
		} else if(oc_out & BOTTOM) {
			t = FIXDIV(fmaxy - fy0, fy1 - fy0);
			x = LERP(fx0, fx1, t);
			y = fmaxy;
		} else if(oc_out & LEFT) {
			t = FIXDIV(fminx - fx0, fx1 - fx0);
			x = fminx;
			y = LERP(fy0, fy1, t);
		} else if(oc_out & RIGHT) {
			t = FIXDIV(fmaxx - fx0, fx1 - fx0);
			x = fmaxx;
			y = LERP(fy0, fy1, t);
		}

		if(oc_out == oc0) {
			fx0 = x;
			fy0 = y;
			oc0 = outcode(fx0 >> 8, fy0 >> 8, minx, miny, maxx, maxy);
		} else {
			fx1 = x;
			fy1 = y;
			oc1 = outcode(fx1 >> 8, fy1 >> 8, minx, miny, maxx, maxy);
		}
	}

	*x0 = fx0 >> 8;
	*y0 = fy0 >> 8;
	*x1 = fx1 >> 8;
	*y1 = fy1 >> 8;
	return 1;
}

void draw_line(uint32_t *fb, int fbwidth, uint32_t col, int x0, int y0, int x1, int y1)
{
	int i, dx, dy, x_inc, y_inc, error;

	fb += y0 * fbwidth + x0;

	dx = x1 - x0;
	dy = y1 - y0;

	if(dx >= 0) {
		x_inc = 1;
	} else {
		x_inc = -1;
		dx = -dx;
	}
	if(dy >= 0) {
		y_inc = fbwidth;
	} else {
		y_inc = -fbwidth;
		dy = -dy;
	}

	if(dx > dy) {
		error = dy * 2 - dx;
		for(i=0; i<=dx; i++) {
			*fb = col;
			if(error >= 0) {
				error -= dx * 2;
				fb += y_inc;
			}
			error += dy * 2;
			fb += x_inc;
		}
	} else {
		error = dx * 2 - dy;
		for(i=0; i<=dy; i++) {
			*fb = col;
			if(error >= 0) {
				error -= dy * 2;
				fb += x_inc;
			}
			error += dx * 2;
			fb += y_inc;
		}
	}
}
