#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <float.h>
#include <direct.h>
#include "inttypes.h"
#include "gfx.h"
#include "keyb.h"
#include "mouse.h"
#include "logger.h"
#include "scene.h"
#include "rayzor.h"
#include "screen.h"
#include "modeller.h"
#include "renderer.h"
#include "scrman.h"
#include "timer.h"

#ifdef __DOS__
#undef USE_ASM_SWAPBUF
#endif

#ifdef USE_ASM_SWAPBUF
// defined in swapbuf.asm
extern "C" void swap_buffers_asm(void *dest, void *src, int xsz, int ysz, int bpp);
#endif

static bool init();
static void cleanup();
static void display();
static void swap_buffers();
static void draw_cursor(uint32_t *buf, int mx, int my);
static void screenshot();
static void handle_keyboard();
static void handle_mouse();
static bool parse_args(int argc, char **argv);
static void sig(int s);

uint32_t *fb_pixels;
int fb_width = 640;
int fb_height = 480;
int fb_bpp = 32;
Scene *scene;

static bool novideo;
static void *fb;
static int rbits, gbits, bbits;
static int rshift, gshift, bshift;
static unsigned int rmask, gmask, bmask;

static bool use_asm_swap = true;
static bool use_mouse;
static int mouse_x, mouse_y;
static bool quit;
static bool cap_shot;

int main(int argc, char **argv)
{
	unsigned long start_msec, msec;
	unsigned long nframes = 0;

	if(!parse_args(argc, argv)) {
		return 1;
	}
	if(!init()) {
		return 1;
	}

	start_msec = get_msec();

	// main loop
	for(;;) {
		handle_keyboard();
		handle_mouse();
		if(quit) break;

		display();
		++nframes;

		if(novideo) break;
	}

	msec = get_msec() - start_msec;

	cleanup();

	printf("Average framerate: %g\n", (float)nframes / ((float)msec / 1000.0f));
	printf("Thank you for using Rayzor!\n");
	return 0;
}

void quit_app()
{
	quit = true;
}

static bool init()
{
	signal(SIGINT, sig);
	signal(SIGSEGV, sig);
	signal(SIGILL, sig);
	signal(SIGFPE, sig);

#ifdef __WATCOM__
	// mask all fpe except invalid op
	_control87(~_EM_INVALID, _MCW_EM);
#endif

	init_timer(128);

	if(!novideo) {
		if(kb_init(32) == -1) {
			fprintf(stderr, "failed to initialize keyboard driver\n");
			return false;
		}

		if(!(fb = set_video_mode(fb_width, fb_height, fb_bpp))) {
			set_text_mode();
			fprintf(stderr, "failed to set video mode: %dx%d %dbpp\n", fb_width, fb_height, fb_bpp);
			return false;
		}
		fb_bpp = get_color_depth();
		get_color_bits(&rbits, &gbits, &bbits);
		get_color_shift(&rshift, &gshift, &bshift);
		get_color_mask(&rmask, &gmask, &bmask);

		printlog("video resolution: %dx%d\n", fb_width, fb_height);
		printlog("bpp: %d (%d %d %d)\n", fb_bpp, rbits, gbits, bbits);
		printlog("shift: %d %d %d\n", rshift, gshift, bshift);
		printlog("mask: %x %x %x\n", rmask, gmask, bmask);

		if(have_mouse()) {
			use_mouse = true;
			set_mouse_limits(0, 0, fb_width - 1, fb_height - 1);
		}
	} else {
		logger_output(stdout);
		printlog("novideo (debug) mode\n");
		fb_bpp = 24;
		rbits = gbits = bbits = 8;
	}

	fb_pixels = new uint32_t[fb_width * fb_height * 4];
	if(!fb_pixels) {
		return false;
	}

	scene = new Scene;

	Sphere *sph = new Sphere;
	sph->mtl.diffuse = Vector3(1.0, 0.3, 0.1);
	sph->mtl.roughness = 0.4;
	scene->add(sph);

	Box *box = new Box;
	box->mtl.diffuse = Vector3(0.1, 0.4, 1.0);
	box->mtl.specular = Vector3(0.3, 0.6, 1.0);
	box->mtl.roughness = 0.2;
	box->mtl.reflect = 0.8;
	box->set_position(Vector3(0, -1.1, 0));
	box->set_scaling(Vector3(4, 0.1, 4));
	scene->add(box);

	Light *lt = new Light;
	lt->set_intensity(0.8);
	lt->set_position(Vector3(-10, 10, 10));
	scene->add(lt);

	Modeller *modeller = new Modeller;
	if(!modeller->init()) {
		return false;
	}
	add_screen(modeller);

	Renderer *renderer = new Renderer;
	if(!renderer->init()) {
		return false;
	}
	add_screen(renderer);

	activate_screen(modeller); // start the modeller screen
	return true;
}

static void cleanup()
{
	delete scene;
	delete [] fb_pixels;

	destroy_screens();

	if(!novideo) {
		set_text_mode();
		kb_shutdown();
	}
}

static void display()
{
	Screen *scr = active_screen();
	if(scr) {
		scr->update();
		scr->draw();
	}

	// draw the mouse cursor
	if(use_mouse) {
		draw_cursor(fb_pixels, mouse_x, mouse_y);
	}

	if(cap_shot) {
		screenshot();
		cap_shot = false;
	}

	if(!novideo) {
		wait_vsync();
#ifdef USE_ASM_SWAPBUF
		swap_buffers_asm(fb, fb_pixels, fb_width, fb_height, fb_bpp);
#else
		swap_buffers();
#endif
	}
}

#define PACK_RGB(r, g, b) \
	((((r) << rshift) & rmask) | \
	 (((g) << gshift) & gmask) | \
	 (((b) << bshift) & bmask))

#define UNPACK_RED(c)	(((c) >> 16) & 0xff)
#define UNPACK_GREEN(c)	(((c) >> 8) & 0xff)
#define UNPACK_BLUE(c)	((c) & 0xff)

static void swap_buffers()
{
	uint32_t *src = fb_pixels;
	int num_pixels = fb_width * fb_height;

	switch(fb_bpp) {
	case 32:
		memcpy(fb, fb_pixels, num_pixels * 4);
		break;

	case 24:
		{
			unsigned char *dest = (unsigned char*)fb;
			for(int i=0; i<num_pixels-1; i++) {
				*((uint32_t*)dest) = *src++;
				dest += 3;
			}
			*dest++ = UNPACK_RED(*src);
			*dest++ = UNPACK_GREEN(*src);
			*dest++ = UNPACK_BLUE(*src);
		}
		break;

	case 16:
		{
			uint16_t *dest = (uint16_t*)fb;
			for(int i=0; i<num_pixels; i++) {
				uint32_t c = *src++;
				unsigned char r = UNPACK_RED(c);
				unsigned char g = UNPACK_GREEN(c);
				unsigned char b = UNPACK_BLUE(c);

				*dest++ = (((r) << 8) & 0xf800) |
					(((g) << 3) & 0x7e0) |
					(((b) >> 3) & 0x1f);
			}
		}
		break;

	default:
		break;
	}
}

static void draw_cursor(uint32_t *buf, int mx, int my)
{
	uint32_t *cptr = buf + my * fb_width + mx;
	int i, cw[2] = {4, 4}, ch[2] = {4, 4};

	if(mx < cw[0]) cw[0] = mx;
	if(my < ch[0]) ch[0] = my;
	if(fb_width - mx < cw[1]) cw[1] = fb_width - mx - 1;
	if(fb_height - my < ch[1]) ch[1] = fb_height - my - 1;

	for(i=1; i<cw[0]; i++) {
		int idx = -i;
		cptr[idx] = 0xffffff;
	}
	for(i=1; i<cw[1]; i++) {
		int idx = i;
		cptr[idx] = 0xffffff;
	}
	for(i=1; i<ch[0]; i++) {
		int idx = -i * fb_width;
		cptr[idx] = 0xffffff;
	}
	for(i=1; i<ch[1]; i++) {
		int idx = i * fb_width;
		cptr[idx] = 0xffffff;
	}
}

#define PPM_COMMENT	"screenshot saved by the rayzor modeller/renderer"
static void screenshot()
{
	static int shotidx = -1;
	FILE *fp;
	char fname[PATH_MAX];

	if(shotidx == -1) {
		DIR *dir;
		struct dirent *dent;

		shotidx = 0;
		if((dir = opendir("."))) {
			while((dent = readdir(dir))) {
				int i, num;
				for(i=0; dent->d_name[i]; i++) {
					fname[i] = tolower(dent->d_name[i]);
				}
				fname[i] = 0;

				if(sscanf(fname, "shot%d.ppm", &num) == 1 && num > shotidx) {
					shotidx = num;
				}
			}
			closedir(dir);
		}
	}

	sprintf(fname, "shot%04d.ppm", ++shotidx);
	if(!(fp = fopen(fname, "wb"))) {
		printlog("failed to save screenshot %s: %s\n", fname, strerror(errno));
		return;
	}

	fprintf(fp, "P6\n# " PPM_COMMENT "\n%d %d\n255\n", fb_width, fb_height);
	for(int i=0; i<fb_width * fb_height; i++) {
		uint32_t c = fb_pixels[i];
		fputc(UNPACK_RED(c), fp);
		fputc(UNPACK_GREEN(c), fp);
		fputc(UNPACK_BLUE(c), fp);
	}

	fclose(fp);
}

static void handle_keyboard()
{
	int key;
	Screen *scr = active_screen();

	if(novideo) return;

	while((key = kb_getkey()) != -1) {
		switch(key) {
		case '`':
			use_asm_swap = !use_asm_swap;
			break;

		case 'q':
		case 'x':
			if(kb_isdown(KB_ALT)) {
				quit_app();
			}
			break;

		case KB_F12:
			cap_shot = true;
			break;

		default:
			break;
		}
		scr->handle_keyboard(key, true);	// TODO also generate release events...
	}
}

static void handle_mouse()
{
	static int prev_mx, prev_my, prev_bnmask, bndiff;
	int mx, my, bnmask;
	Screen *scr = active_screen();

	if(!use_mouse || novideo) return;

	bnmask = read_mouse(&mx, &my);
	if(scr && (bndiff = bnmask ^ prev_bnmask)) {
		for(int i=0; i<8; i++) {
			int bit = 1 << i;
			if(bndiff & bit) {
				scr->handle_mbutton(i, bnmask & bit, mx, my);
			}
		}
	}
	prev_bnmask = bnmask;

	if(scr && (mx != prev_mx || my != prev_my)) {
		scr->handle_mmotion(mx, my);
	}
	prev_mx = mx;
	prev_my = my;

	mouse_x = mx;
	mouse_y = my;
}


static struct {
	int opt;
	const char *lopt;
	const char *desc;
} options[] = {
	{'s', "size", "resolution <xres>x<yres>[:bpp]"},
	{'n', "novid", "don't switch video mode (for debugging)"},
	{'h', "help", "print usage information and exit"},
	{-1, 0, 0}
};

static void print_usage(const char *argv0)
{
	printf("%s usage\n", argv0);
	for(int i=0; options[i].opt != -1; i++) {
		printf("  -%c, -%s: %s\n", options[i].opt, options[i].lopt, options[i].desc);
	}
	exit(0);
}

static bool parse_args(int argc, char **argv)
{
	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			int opt = -1;

			for(int j=0; options[j].opt != -1; j++) {
				if(argv[i][2] == 0) {
					if(argv[i][1] == options[j].opt) {
						opt = options[j].opt;
						break;
					}
				} else {
					if(strcmp(argv[i] + 1, options[j].lopt) == 0) {
						opt = options[j].opt;
						break;
					}
				}
			}

			switch(opt) {
			case 's':
				if(sscanf(argv[++i], "%dx%d:%d", &fb_width, &fb_height, &fb_bpp) < 2) {
					fprintf(stderr, "%s must be followed by a resolution: WxH\n", argv[i - 1]);
					return false;
				}
				break;

			case 'n':
				novideo = true;
				break;

			case 'h':
				print_usage(argv[0]);	// doesn't return
				break;

			default:
				fprintf(stderr, "unknown option: %s\n", argv[i]);
				return false;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return false;
		}
	}
	return true;
}

static void sig(int s)
{
	cleanup();
	fprintf(stderr, "signal caught: %d\n", s);

#ifdef __WATCOM__
	if(s == SIGFPE) {
		unsigned int st = _status87();
		fprintf(stderr, "fpu status: %x\n", st);
	}
#endif

	exit(1);
}
