#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "min3d.h"
#include "m3dimpl.h"
#include "logger.h"

#ifndef M_PI
#define M_PI	3.141592653
#endif

struct min3d_context *m3dctx;

int m3d_init(void)
{
	if(!(m3dctx = malloc(sizeof *m3dctx))) {
		return -1;
	}
	memset(m3dctx, 0, sizeof *m3dctx);

	m3d_matrix_mode(M3D_PROJECTION);
	m3d_load_identity();
	m3d_matrix_mode(M3D_MODELVIEW);
	m3d_load_identity();

	m3d_color(1, 1, 1);
	return 0;
}

void m3d_shutdown(void)
{
	free(m3dctx);
}

void m3d_set_buffers(struct m3d_image *cbuf, uint16_t *zbuf)
{
	m3dctx->cbuf = cbuf;
	m3dctx->zbuf = zbuf;

	m3d_viewport(0, 0, cbuf->xsz, cbuf->ysz);
}

void m3d_clear_color(float r, float g, float b)
{
	m3dctx->clear_color[0] = (int)((r > 1.0 ? 1.0 : r) * 255.0);
	m3dctx->clear_color[1] = (int)((g > 1.0 ? 1.0 : g) * 255.0);
	m3dctx->clear_color[2] = (int)((b > 1.0 ? 1.0 : b) * 255.0);
}

void m3d_clear(unsigned int bmask)
{
	int num_pixels = m3dctx->cbuf->xsz * m3dctx->cbuf->ysz;
	if(bmask & M3D_COLOR_BUFFER_BIT) {
		memset(m3dctx->cbuf->pixels, 0, num_pixels * 4);
		/*
		int i;
		unsigned char *ptr = m3dctx->cbuf->pixels;
		unsigned char r = m3dctx->clear_color[0];
		unsigned char g = m3dctx->clear_color[1];
		unsigned char b = m3dctx->clear_color[2];
		for(i=0; i<num_pixels; i++) {
			*ptr++ = r;
			*ptr++ = g;
			*ptr++ = b;
		}*/
	}
	if(bmask & M3D_DEPTH_BUFFER_BIT) {
		memset(m3dctx->zbuf, 0xff, num_pixels * sizeof *m3dctx->zbuf);
	}
}


void m3d_enable(int bit)
{
	m3dctx->state |= (1 << bit);
}

void m3d_disable(int bit)
{
	m3dctx->state &= ~(1 << bit);
}

void m3d_viewport(int x, int y, int xsz, int ysz)
{
	m3dctx->vport[0] = x;
	m3dctx->vport[1] = y;
	m3dctx->vport[2] = xsz;
	m3dctx->vport[3] = ysz;
}


/* matrix stack */
void m3d_matrix_mode(int mode)
{
	m3dctx->mmode = mode;
}

void m3d_push_matrix(void)
{
	int mm = m3dctx->mmode;
	int top = m3dctx->mstack[mm].top;
	if(top < MSTACK_SIZE) {
		float *cur = m3dctx->mstack[mm].m[top++];
		memcpy(m3dctx->mstack[mm].m[top], cur, 16 * sizeof *cur);
		m3dctx->mstack[mm].top = top;
	}
}

void m3d_pop_matrix(void)
{
	int mm = m3dctx->mmode;
	if(m3dctx->mstack[mm].top > 0) {
		--m3dctx->mstack[mm].top;
	}
}

void m3d_load_identity(void)
{
	static const float mid[] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
	m3d_load_matrix(mid);
}

void m3d_load_matrix(const float *m)
{
	int top = m3dctx->mstack[m3dctx->mmode].top;
	memcpy(m3dctx->mstack[m3dctx->mmode].m[top], m, 16 * sizeof *m);
}

void m3d_load_transpose_matrix(const float *m)
{
	int i, j;
	int top = m3dctx->mstack[m3dctx->mmode].top;
	float *dest = m3dctx->mstack[m3dctx->mmode].m[top];

	for(i=0; i<4; i++) {
		for(j=0; j<4; j++) {
			*dest++ = m[j * 4 + i];
		}
	}
}

#define M(i,j)	(((i) << 2) + (j))
void m3d_mult_matrix(const float *m2)
{
	int i, j, top = m3dctx->mstack[m3dctx->mmode].top;
	float m1[16];
	float *dest = m3dctx->mstack[m3dctx->mmode].m[top];

	memcpy(m1, dest, sizeof m1);

	for(i=0; i<4; i++) {
		for(j=0; j<4; j++) {
			*dest++ = m1[M(0,j)] * m2[M(i,0)] +
				m1[M(1,j)] * m2[M(i,1)] +
				m1[M(2,j)] * m2[M(i,2)] +
				m1[M(3,j)] * m2[M(i,3)];
		}
	}
}

void m3d_mult_transpose_matrix(const float *m2)
{
	int i, j, top = m3dctx->mstack[m3dctx->mmode].top;
	float m1[16];
	float *dest = m3dctx->mstack[m3dctx->mmode].m[top];

	memcpy(m1, dest, sizeof m1);

	for(i=0; i<4; i++) {
		for(j=0; j<4; j++) {
			*dest++ = m1[M(0,j)] * m2[M(0,i)] +
				m1[M(1,j)] * m2[M(1,i)] +
				m1[M(2,j)] * m2[M(2,i)] +
				m1[M(3,j)] * m2[M(3,i)];
		}
	}
}

void m3d_translate(float x, float y, float z)
{
	float m[] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
	m[12] = x;
	m[13] = y;
	m[14] = z;
	m3d_mult_matrix(m);
}

void m3d_rotate(float deg, float x, float y, float z)
{
	float xform[] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};

	float angle = M_PI * deg / 180.0f;
	float sina = sin(angle);
	float cosa = cos(angle);
	float one_minus_cosa = 1.0f - cosa;
	float nxsq = x * x;
	float nysq = y * y;
	float nzsq = z * z;

	xform[0] = nxsq + (1.0f - nxsq) * cosa;
	xform[4] = x * y * one_minus_cosa - z * sina;
	xform[8] = x * z * one_minus_cosa + y * sina;
	xform[1] = x * y * one_minus_cosa + z * sina;
	xform[5] = nysq + (1.0 - nysq) * cosa;
	xform[9] = y * z * one_minus_cosa - x * sina;
	xform[2] = x * z * one_minus_cosa - y * sina;
	xform[6] = y * z * one_minus_cosa + x * sina;
	xform[10] = nzsq + (1.0 - nzsq) * cosa;

	m3d_mult_matrix(xform);
}

void m3d_scale(float x, float y, float z)
{
	static float m[] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
	m[0] = x;
	m[5] = y;
	m[10] = z;
	m3d_mult_matrix(m);
}

void m3d_frustum(float left, float right, float bottom, float top, float nr, float fr)
{
	float xform[] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};

	float dx = right - left;
	float dy = top - bottom;
	float dz = fr - nr;

	float a = (right + left) / dx;
	float b = (top + bottom) / dy;
	float c = -(fr + nr) / dz;
	float d = -2.0 * fr * nr / dz;

	xform[0] = 2.0 * nr / dx;
	xform[5] = 2.0 * nr / dy;
	xform[8] = a;
	xform[9] = b;
	xform[10] = c;
	xform[11] = -1.0f;
	xform[14] = d;

	m3d_mult_matrix(xform);
}

void m3d_perspective(float vfov, float aspect, float nr, float fr)
{
	float vfov_rad = M_PI * vfov / 180.0;
	float x = nr * tan(vfov_rad / 2.0);
	m3d_frustum(-aspect * x, aspect * x, -x, x, nr, fr);
}

static void xform4(float *mat, float *vec)
{
	float x = mat[0] * vec[0] + mat[4] * vec[1] + mat[8] * vec[2] + mat[12];
	float y = mat[1] * vec[0] + mat[5] * vec[1] + mat[9] * vec[2] + mat[13];
	float z = mat[2] * vec[0] + mat[6] * vec[1] + mat[10] * vec[2] + mat[14];
	float w = mat[3] * vec[0] + mat[7] * vec[1] + mat[11] * vec[2] + mat[15];

	vec[0] = x;
	vec[1] = y;
	vec[2] = z;
	vec[3] = w;
}

static int proc_prim(int prim, struct min3d_vertex *res, struct min3d_vertex *v)
{
	int i;
	int vcount = prim;
	int mvtop, ptop;
	float *mvmat, *pmat;
	int *vport = m3dctx->vport;

	mvtop = m3dctx->mstack[M3D_MODELVIEW].top;
	mvmat = m3dctx->mstack[M3D_MODELVIEW].m[mvtop];
	ptop = m3dctx->mstack[M3D_PROJECTION].top;
	pmat = m3dctx->mstack[M3D_PROJECTION].m[ptop];

	/* transform to view space */
	for(i=0; i<vcount; i++) {
		res[i] = v[i];
		xform4(mvmat, res[i].pos);
		/* TODO: normal */
	}

	/* TODO: lighting */

	/* project */
	for(i=0; i<vcount; i++) {
		xform4(pmat, res[i].pos);
	}

	/* clip */
	switch(prim) {
	case M3D_POINTS:
		{
			float w = res[0].pos[3];
			if(res[0].pos[2] < -w || res[0].pos[2] >= w ||
					res[0].pos[0] / w < -1 || res[0].pos[0] / w >= 1 ||
					res[0].pos[1] / w < -1 || res[0].pos[1] / w >= 1) {
				vcount = 0;
			}
		}
		break;

	default:
		break;	/* TODO */
	}

	/* perspective division & viewport */
	for(i=0; i<vcount; i++) {
		res[i].pos[0] /= res[i].pos[3];
		res[i].pos[1] /= res[i].pos[3];
		res[i].pos[2] /= res[i].pos[3];

		res[i].pos[0] = (res[i].pos[0] * 0.5 + 0.5) * vport[2] + vport[0];
		res[i].pos[1] = (-res[i].pos[1] * 0.5 + 0.5) * vport[3] + vport[1];
	}
	return vcount;
}

/* drawing */
void m3d_vertex_array(const float *varr)
{
	m3dctx->vert_array = (float*)varr;
}

void m3d_normal_array(const float *narr)
{
	m3dctx->norm_array = (float*)narr;
}

void m3d_color_array(const float *carr)
{
	m3dctx->col_array = (float*)carr;
}

void m3d_texcoord_array(const float *tcarr)
{
	m3dctx->tc_array = (float*)tcarr;
}


void m3d_draw(int prim, int vcount)
{
	int i;
	struct min3d_vertex v[4];
	struct min3d_vertex resv[16];
	const float *varr = m3dctx->vert_array;
	const float *carr = m3dctx->col_array;

	if(!varr) return;

	for(i=0; i<vcount; i++) {
		int r, g, b;
		int idx = i % prim;

		v[idx].pos[0] = *varr++;
		v[idx].pos[1] = *varr++;
		v[idx].pos[2] = *varr++;
		v[idx].pos[3] = 1.0;
		r = (carr ? *carr++ : m3dctx->im_color[0]) * 255.0;
		g = (carr ? *carr++ : m3dctx->im_color[1]) * 255.0;
		b = (carr ? *carr++ : m3dctx->im_color[2]) * 255.0;
		v[idx].color = (r << 16) | (g << 8) | b;

		if(idx == prim - 1) {
			int resnum = proc_prim(prim, resv, v);
			switch(resnum) {
			case 1:
				m3d_draw_point(resv);
				break;

			case 2:
				m3d_draw_line(resv);
				break;

			default:
				m3d_draw_poly(resv, resnum);
			}
		}
	}
}

void m3d_draw_indexed(int prim, const unsigned int *idxarr, int icount)
{
	int i, vcount = prim;
	struct min3d_vertex v[4];
	struct min3d_vertex resv[16];
	const float *varr = m3dctx->vert_array;
	const float *carr = m3dctx->col_array;

	if(!varr) return;

	for(i=0; i<icount; i++) {
		int r, g, b;
		int vnum = i % vcount;
		int index = idxarr[i];

		v[vnum].pos[0] = varr[index * 3];
		v[vnum].pos[1] = varr[index * 3 + 1];
		v[vnum].pos[2] = varr[index * 3 + 2];
		v[vnum].pos[3] = 1.0;
		r = (carr ? carr[index * 3] : m3dctx->im_color[0]) * 255.0;
		g = (carr ? carr[index * 3 + 1] : m3dctx->im_color[1]) * 255.0;
		b = (carr ? carr[index * 3 + 2] : m3dctx->im_color[2]) * 255.0;
		v[vnum].color = (r << 16) | (g << 8) | b;

		if(vnum == vcount - 1) {
			int resnum = proc_prim(prim, resv, v);
			switch(resnum) {
			case 1:
				m3d_draw_point(resv);
				break;

			case 2:
				m3d_draw_line(resv);
				break;

			default:
				m3d_draw_poly(resv, resnum);
			}
		}
	}
}

void m3d_begin(int prim)
{
	m3dctx->im_prim = prim;
	m3dctx->im_idx = 0;

	m3dctx->vert_array = m3dctx->im_varr;
	m3dctx->norm_array = 0;
	m3dctx->col_array = 0;
	m3dctx->tc_array = 0;
}

void m3d_end(void)
{
}

void m3d_vertex(float x, float y, float z)
{
	int nverts = m3dctx->im_prim;
	int idx = m3dctx->im_idx;
	float *v = m3dctx->vert_array + idx * 3;

	v[0] = x;
	v[1] = y;
	v[2] = z;

	if(m3dctx->norm_array) {
		float *ptr = m3dctx->im_narr + idx * 3;
		ptr[0] = m3dctx->im_normal[0];
		ptr[1] = m3dctx->im_normal[1];
		ptr[2] = m3dctx->im_normal[2];
	}
	if(m3dctx->col_array) {
		float *ptr = m3dctx->im_carr + idx * 3;
		ptr[0] = m3dctx->im_color[0];
		ptr[1] = m3dctx->im_color[1];
		ptr[2] = m3dctx->im_color[2];
	}
	if(m3dctx->tc_array) {
		float *ptr = m3dctx->im_texcoord + idx * 2;
		ptr[0] = m3dctx->im_texcoord[0];
		ptr[1] = m3dctx->im_texcoord[1];
	}

	if(++idx == nverts) {
		m3d_draw(m3dctx->im_prim, nverts);
		idx = 0;
	}

	m3dctx->im_idx = idx;
}

void m3d_normal(float x, float y, float z)
{
	m3dctx->im_normal[0] = x;
	m3dctx->im_normal[1] = y;
	m3dctx->im_normal[2] = z;
}

void m3d_color(float x, float y, float z)
{
	m3dctx->im_color[0] = x;
	m3dctx->im_color[1] = y;
	m3dctx->im_color[2] = z;
}

void m3d_texcoord(float x, float y)
{
	m3dctx->im_texcoord[0] = x;
	m3dctx->im_texcoord[1] = y;
}
