#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "modeller.h"
#include "min3d.h"
#include "rayzor.h"
#include "scene.h"
#include "keyb.h"
#include "scrman.h"
#include "logger.h"
#include "timer.h"

struct ModellerImpl {
	int mx, my;
	float cam_theta, cam_phi, cam_dist, cam_fov;
	Camera *viewport_cam;

	struct m3d_image rbuf;

	bool bnstate[8];
	int prev_x, prev_y;

	float orig_theta;
	bool screensaver;
};

static void draw_grid(float size, float spacing);

Modeller::Modeller()
{
	set_name("modeller");
}

Modeller::~Modeller()
{
	shutdown();
}

bool Modeller::init()
{
	mod = new ModellerImpl;
	memset(mod, 0, sizeof *mod);

	mod->viewport_cam = new Camera;

	mod->cam_phi = 25;
	mod->cam_dist = 5;
	mod->cam_fov = 50.0;

	m3d_init();
	mod->rbuf.pixels = fb_pixels;
	mod->rbuf.xsz = fb_width;
	mod->rbuf.ysz = fb_height;
	m3d_set_buffers(&mod->rbuf, 0);

	m3d_matrix_mode(M3D_PROJECTION);
	m3d_load_identity();
	m3d_perspective(mod->cam_fov, (float)fb_width / (float)fb_height, 0.5, 500.0);

	m3d_enable(M3D_CULL_FACE);
	return true;
}

void Modeller::shutdown()
{
	if(mod) {
		m3d_shutdown();

		delete mod;
		mod = 0;
	}
}

void Modeller::draw() const
{
	if(mod->screensaver) {
		mod->cam_theta = get_msec() / 100.0f;
	}

	m3d_clear(M3D_COLOR_BUFFER_BIT);

	m3d_matrix_mode(M3D_MODELVIEW);
	m3d_load_identity();
	m3d_translate(0, 0, -mod->cam_dist);
	m3d_rotate(mod->cam_phi, 1, 0, 0);
	m3d_rotate(mod->cam_theta, 0, 1, 0);

	draw_grid(10.0, 1.0);

	scene->draw();
}


static void draw_grid(float size, float spacing)
{
	int num_lines = size / spacing;
	float dist = size / 2.0;

	m3d_disable(M3D_LIGHTING);

	m3d_begin(M3D_LINES);
	m3d_color(0.4, 0.4, 0.4);

	float x = -dist;
	for(int i=0; i<=num_lines; i++) {
		if(i != num_lines / 2) {
			m3d_vertex(-dist, 0, x);
			m3d_vertex(dist, 0, x);
			m3d_vertex(x, 0, -dist);
			m3d_vertex(x, 0, dist);
		}
		x += spacing;
	}
	m3d_end();

	m3d_begin(M3D_LINES);
	m3d_color(0.8, 0, 0);
	m3d_vertex(-dist, 0, 0);
	m3d_vertex(dist, 0, 0);
	m3d_color(0.1, 0.3, 0.8);
	m3d_vertex(0, 0, -dist);
	m3d_vertex(0, 0, dist);
	m3d_end();
}


void Modeller::handle_keyboard(int key, bool press)
{
	if(press) {
		switch(key) {
		case 'q':
			quit_app();
			break;

		case 27:
			scene->clear_selection();
			break;

		case '\t':
			{
				int s = scene->get_selection();
				if(s >= 0) {
					s = (s + 1) % scene->get_node_count();
					scene->clear_selection();
				} else {
					s = 0;
				}
				scene->select(s);
			}
			break;

		case 'r':
		case 'p':
			if(kb_isdown(KB_ALT) || kb_isdown(KB_CTRL)) {
		case KB_F5:
		case KB_F6:
				Screen *rs = get_screen("renderer");
				if(rs) {
					activate_screen(rs);

					if(key == 'r' || key == KB_F5) {
						// start a rendering, and make sure there is a camera
						if(!scene->get_active_camera()) {
							scene->set_active_camera(mod->viewport_cam);
						}
						Vector3 dir;
						float theta = -DEG2RAD(mod->cam_theta);
						float phi = DEG2RAD(mod->cam_phi);
						dir.x = sin(theta) * cos(phi) * mod->cam_dist;
						dir.y = sin(phi) * mod->cam_dist;
						dir.z = cos(theta) * cos(phi) * mod->cam_dist;
						mod->viewport_cam->set_position(dir);
						mod->viewport_cam->set_target(Vector3(0, 0, 0));
						mod->viewport_cam->set_fov(DEG2RAD(mod->cam_fov));

						rs->message(message_atom("start"));
					}
				} else {
					printlog("failed to find renderer screen!\n");
				}
			}
			break;

		case KB_F1:
			mod->screensaver = !mod->screensaver;
			if(mod->screensaver) {
				mod->orig_theta = mod->cam_theta;
			} else {
				mod->cam_theta = mod->orig_theta;
			}
			break;

		default:
			break;
		}
	}
}

void Modeller::handle_mbutton(int bn, bool press, int x, int y)
{
	mod->bnstate[bn] = press;
	mod->prev_x = x;
	mod->prev_y = y;
}

void Modeller::handle_mmotion(int x, int y)
{
	int dx = x - mod->prev_x;
	int dy = y - mod->prev_y;
	mod->prev_x = x;
	mod->prev_y = y;

	if(kb_isdown(KB_ALT) || kb_isdown(KB_CTRL)) {
		if(mod->bnstate[0]) {
			mod->cam_theta += dx * 0.5;
			mod->cam_phi += dy * 0.5;

			if(mod->cam_phi < -90) mod->cam_phi = -90;
			if(mod->cam_phi > 90) mod->cam_phi = 90;
		}
		if(mod->bnstate[1]) {
			mod->cam_dist += dy * 0.1;
			if(mod->cam_dist < 0) mod->cam_dist = 0;
		}
	}
}
