#include <math.h>
#include "object.h"
#include "vmath.h"
#include "min3d.h"
#include "logger.h"

Object::Object()
{
	type = NODE_OBJECT;
}

Object::~Object()
{
}

Vector3 Object::hit_normal(const RayHit &hit) const
{
	return Vector3(0, 0, 0);
}

// ---- sphere ----
Sphere::Sphere()
{
}

Sphere::~Sphere()
{
}

#define USUB	12
#define VSUB	6

void Sphere::draw(bool emph) const
{
	static Vector3 *varr;
	static unsigned int *iarr;
	static int num_verts, num_indices;
	if(!varr) {
		int i, j;
		int uverts = USUB;
		int vverts = VSUB + 1;

		num_verts = uverts * vverts;
		varr = new Vector3[num_verts];

		Vector3 *vptr = varr;
		for(i=0; i<vverts; i++) {
			float v = (float)i / (float)VSUB;
			float phi = v * M_PI;
			for(j=0; j<uverts; j++) {
				float u = (float)j / (float)uverts;
				float theta = u * M_PI * 2.0;

				float x = sin(theta) * sin(phi);
				float y = -cos(phi);
				float z = cos(theta) * sin(phi);

				*vptr++ = Vector3(x, y, z);
			}
		}

		num_indices = USUB * VSUB * 4;
		iarr = new unsigned int[num_indices];

		unsigned int *iptr = iarr;
		for(i=0; i<VSUB; i++) {
			for(j=0; j<USUB; j++) {
				iptr[0] = i * uverts + j;
				iptr[1] = i * uverts + ((j + 1) % uverts);
				iptr[2] = iptr[1] + uverts;
				iptr[3] = iptr[0] + uverts;
				iptr += 4;
			}
		}

		printlog("created sphere mesh\n");
		printlog("  vertices: %d (%d slices, %d stacks)\n", num_verts, uverts, vverts);
		printlog("  quads: %d (%d indices, %d usub, %d vsub)\n", USUB * VSUB, num_indices, USUB, VSUB);
	}

	pre_draw();
	SceneNode::draw(emph);

	m3d_vertex_array(&varr->x);
	m3d_draw_indexed(M3D_QUADS, iarr, num_indices);
	m3d_vertex_array(0);

	post_draw();
}

bool Sphere::intersect(const Ray &wray, RayHit *hit) const
{
	Ray ray = transform(get_inv_matrix(), wray);

	// assumes center is 0,0,0, and radius is 1
	float a = dot(ray.dir, ray.dir);
	float b = 2.0 * ray.dir.x * ray.origin.x +
		2.0 * ray.dir.y * ray.origin.y +
		2.0 * ray.dir.z * ray.origin.z;
	float c = dot(ray.origin, ray.origin) - 1.0;

	float discr = b * b - 4.0 * a * c;
	if(discr < 1e-4)
		return false;

	float sqrt_discr = sqrt(discr);
	float t0 = (-b + sqrt_discr) / (2.0 * a);
	float t1 = (-b - sqrt_discr) / (2.0 * a);

	if(t0 < 1e-4)
		t0 = t1;
	if(t1 < 1e-4)
		t1 = t0;

	float t = t0 < t1 ? t0 : t1;
	if(t < 1e-4)
		return false;

	if(hit) {
		hit->ray = wray;
		hit->lray = ray;
		hit->dist = t;
		hit->obj = this;
		hit->subobj = 0;
	}
	return true;
}

Vector3 Sphere::hit_normal(const RayHit &hit) const
{
	return hit.lray.origin + hit.lray.dir * hit.dist;
}

// ---- box ----

Box::Box()
{
}

Box::~Box()
{
}

/*
     3--------2
    /.        .\
   0------------1
   | 7--------6 |
   |/          \|
   4------------5

 */
void Box::draw(bool emph) const
{
	static const float verts[] = {
		-1, 1, 1,
		1, 1, 1,
		1, 1, -1,
		-1, 1, -1,
		-1, -1, 1,
		1, -1, 1,
		1, -1, -1,
		-1, -1, -1
	};
	static const unsigned int indices[] = {
		0, 1, 2, 3,		// top
		4, 7, 6, 5,		// bottom
		0, 4, 5, 1,		// front
		5, 6, 2, 1,		// right
		6, 7, 3, 2,		// back
		7, 4, 0, 3		// left
	};

	pre_draw();
	SceneNode::draw(emph);

	m3d_vertex_array(verts);
	m3d_draw_indexed(M3D_QUADS, indices, sizeof indices / sizeof *indices);
	m3d_vertex_array(0);

	post_draw();
}

bool Box::intersect(const Ray &wray, RayHit *hit) const
{
	Ray ray = transform(get_inv_matrix(), wray);

	Vector3 param[2] = { Vector3(-1, -1, -1), Vector3(1, 1, 1) };
	Vector3 inv_dir(1.0 / ray.dir.x, 1.0 / ray.dir.y, 1.0 / ray.dir.z);
	int sign[3] = {inv_dir.x < 0, inv_dir.y < 0, inv_dir.z < 0};

	float tmin = (param[sign[0]].x - ray.origin.x) * inv_dir.x;
	float tmax = (param[1 - sign[0]].x - ray.origin.x) * inv_dir.x;
	float tymin = (param[sign[1]].y - ray.origin.y) * inv_dir.y;
	float tymax = (param[1 - sign[1]].y - ray.origin.y) * inv_dir.y;

	if(tmin > tymax || tymin > tmax) {
		return false;
	}
	if(tymin > tmin) tmin = tymin;
	if(tymax < tmax) tmax = tymax;

	float tzmin = (param[sign[2]].z - ray.origin.z) * inv_dir.z;
	float tzmax = (param[1 - sign[2]].z - ray.origin.z) * inv_dir.z;

	if(tmin > tzmax || tzmin > tmax) {
		return false;
	}
	if(tzmin > tmin) tmin = tzmin;
	if(tzmax < tmax) tmax = tzmax;

	float t = tmin < 1e-4 ? tmax : tmin;
	if(t >= 1e-4) {
		if(hit) {
			hit->dist = t;
			hit->ray = wray;
			hit->lray = ray;
			hit->obj = this;
			hit->subobj = 0;
		}
		return true;
	}
	return false;
}

Vector3 Box::hit_normal(const RayHit &hit) const
{
	Vector3 lpt = hit.lray.origin + hit.lray.dir * hit.dist;

	if(fabs(lpt.x) > fabs(lpt.y) && fabs(lpt.x) > fabs(lpt.z)) {
		return Vector3(lpt.x, 0, 0);
	}
	if(fabs(lpt.y) > fabs(lpt.z)) {
		return Vector3(0, lpt.y, 0);
	}
	return Vector3(0, 0, lpt.z);
}
