#ifndef QUATERNION_H_
#define	QUATERNION_H_

class Quat;

inline Quat operator *(const Quat &a, const Quat &b);

class Quat{
public:
	float x, y, z, w;

	Quat() : x(0), y(0), z(0), w(1) {}
	Quat(float xx, float yy, float zz, float ww) : x(xx), y(yy), z(zz), w(ww) {}

	void set_identity()
	{
		x = y = z = 0.0f;
		w = 1.0f;
	}

	Quat conjugate() const
	{
		return Quat(-x, -y, -z, w);
	}

	float length() const
	{
		return (float)sqrt(x * x + y * y + z * z + w * w);
	}

	float length_sq() const
	{
		return x * x + y * y + z * z + w * w;
	}

	void normalize()
	{
		float len = length();
		if(len != 0.0) {
			x /= len;
			y /= len;
			z /= len;
			w /= len;
		}
	}

	Quat inverse() const
	{
		Quat inv = conjugate();
		float len_sq = length_sq();
		if(len_sq != 0.0) {
			inv.x /= len_sq;
			inv.y /= len_sq;
			inv.z /= len_sq;
			inv.w /= len_sq;
		}
		return inv;
	}

	void set_rotation(float angle, float axis_x, float axis_y, float axis_z)
	{
		float half_angle = angle * 0.5;
		float sin_half = sin(half_angle);

		w = cos(half_angle);
		x = axis_x * sin_half;
		y = axis_y * sin_half;
		z = axis_z * sin_half;
	}

	void rotate(float angle, float x, float y, float z)
	{
		Quat q;
		q.set_rotation(angle, x, y, z);
		*this = *this * q;
	}

	void rotate(const Quat &q)
	{
		*this = q * *this * q.conjugate();
	}

	Matrix4x4 get_matrix() const
	{
		return Matrix4x4(
			1.0 - 2.0 * y*y - 2.0 * z*z,	2.0 * x * y - 2.0 * w * z,		2.0 * z * x + 2.0 * w * y, 0,
			2.0 * x * y + 2.0 * w * z,		1.0 - 2.0 * x*x - 2.0 * z*z,	2.0 * y * z - 2.0 * w * x, 0,
			2.0 * z * x - 2.0 * w * y,		2.0 * y * z + 2.0 * w * x,		1.0 - 2.0 * x*x - 2.0 * y*y, 0,
			0, 0, 0, 1);
	}
};

inline Quat operator *(const Quat &a, const Quat &b)
{
	float dot = a.x * b.x + a.y * b.y + a.z * b.z;
	float cross_x = a.y * b.z - a.z * b.y;
	float cross_y = a.z * b.x - a.x * b.z;
	float cross_z = a.x * b.y - a.y * b.x;

	float w = a.w * b.w - dot;
	float x = a.x * b.w + b.x * a.w + cross_x;
	float y = a.y * b.w + b.y * a.w + cross_y;
	float z = a.z * b.w + b.z * a.w + cross_z;

	return Quat(x, y, z, w);
}

#endif	// QUATERNION_H_
