#include <assert.h>
#include <float.h>
#include "raytrace.h"
#include "rayzor.h"
#include "scene.h"
#include "logger.h"

Vector3 ray_trace(const Ray &ray, int iter)
{
	RayHit hit;
	if(!scene->intersect(ray, &hit)) {
		return scene->get_background(ray);
	}

	return shade(hit, iter);
}

static inline float positive(float x)
{
	return x < 0.0f ? 0.0f : x;
}

Vector3 shade(const RayHit &hit, int iter)
{
	const Material &mtl = hit.obj->mtl;

	Vector3 pos = hit.ray.origin + hit.ray.dir * hit.dist;
	Vector3 norm = hit.obj->hit_normal(hit);
	norm = normalize(transform(normal_matrix(hit.obj->get_matrix()), norm));
	Vector3 vdir = -normalize(hit.ray.dir);

	float ior = mtl.ior;

	if(dot(norm, hit.ray.dir) > 0.0) {
		norm = -norm;
		ior = 1.0 / mtl.ior;
	}

	Vector3 color = scene->get_ambient();

	// for each light, calculate local illumination
	int num_lights = scene->get_light_count();
	for(int i=0; i<num_lights; i++) {
		const Light *lt = scene->get_light(i);
		Vector3 ldir = lt->get_position() - pos;
		Vector3 lcol = lt->get_color(pos);

		RayHit hit;
		if(!scene->intersect(Ray(pos, ldir), &hit) || hit.dist > 1.0) {
			// if we can see the light, calculate and add its contribution
			ldir.normalize();
			float ndotl = positive(dot(norm, ldir));
			Vector3 diffuse = mtl.diffuse * lcol * ndotl;

			Vector3 hdir = normalize(ldir + vdir);
			float ndoth = positive(dot(norm, hdir));
			Vector3 specular = mtl.specular * lcol * pow(ndoth, mtl.roughness * 128.0);

			color = color + lerp(specular, diffuse, mtl.roughness);
		}
	}

	if(mtl.reflect > 1e-4 && iter < 6) {
		Ray reflray(pos, reflect(vdir, norm));

		Vector3 rcol = ray_trace(reflray, iter + 1) * mtl.specular * mtl.reflect;
		color = color + rcol;
	}

	if(mtl.refract > 1e-4 && iter < 6) {
		Ray refrray(pos, refract(vdir, norm,  ior));

		Vector3 rcol = ray_trace(refrray, iter + 1) * mtl.specular * mtl.refract;
		color = color + rcol;
	}

	return color;
}
