#include <string.h>
#include <stdarg.h>
#include "renderer.h"
#include "rayzor.h"
#include "scene.h"
#include "keyb.h"
#include "scrman.h"
#include "timer.h"
#include "logger.h"
#include "raytrace.h"

enum {
	MSG_START,
	MSG_GAMMA,
	MSG_USE_GAMMA,

	NUM_MESSAGES
};

struct RendererImpl {
	bool in_progress;
	int cur_x, cur_y;	// current pixel being rendered

	uint32_t *pixels;

	float gamma, inv_gamma;
	bool use_gamma;

	MsgAtom msg[NUM_MESSAGES];
};


Renderer::Renderer()
{
	set_name("renderer");
}

bool Renderer::init()
{
	rend = new RendererImpl;
	memset(rend, 0, sizeof *rend);

	rend->pixels = new uint32_t[fb_width * fb_height];
	if(!rend->pixels) {
		return false;
	}
	memset(rend->pixels, 0, fb_width * fb_height * 4);

	rend->gamma = 1.8;
	rend->inv_gamma = 1.0 / rend->gamma;
	rend->use_gamma = true;

	rend->msg[MSG_START] = message_atom("start");
	rend->msg[MSG_GAMMA] = message_atom("gamma");
	rend->msg[MSG_USE_GAMMA] = message_atom("use-gamma");

	return true;
}

void Renderer::shutdown()
{
	if(rend) {
		delete [] rend->pixels;
		delete rend;
		rend = 0;
	}
}

void Renderer::draw() const
{
	Camera *cam = scene->get_active_camera();

	// if a rendering is not in progress, just show the last buffer
	if(!rend->in_progress || !cam) {
		memcpy(fb_pixels, rend->pixels, fb_width * fb_height * 4);
		return;
	}

	// render for approximately half a second, then show and return
	unsigned long start = get_msec();
	uint32_t *dest = rend->pixels + rend->cur_y * fb_width + rend->cur_x;

	for(;;) {
		Ray ray = cam->get_primary_ray(rend->cur_x, rend->cur_y);
		Vector3 color = ray_trace(ray);

		// gamma-correct before storing the pixel
		if(rend->use_gamma) {
			color.x = pow(color.x, rend->inv_gamma);
			color.y = pow(color.y, rend->inv_gamma);
			color.z = pow(color.z, rend->inv_gamma);
		}

		int r = color.x > 1.0 ? 255 : (int)(color.x * 255.0);
		int g = color.y > 1.0 ? 255 : (int)(color.y * 255.0);
		int b = color.z > 1.0 ? 255 : (int)(color.z * 255.0);

		*dest++ = (r << 16) | (g << 8) | b;

		if(++rend->cur_x >= fb_width) {
			rend->cur_x = 0;
			if(++rend->cur_y >= fb_height) {
				rend->in_progress = false;
				printlog("done!\n");
				break;
			}
		}

		if(get_msec() - start >= 500) {
			break;
		}
	}

	memcpy(fb_pixels, rend->pixels, fb_width * fb_height * 4);
}

void Renderer::handle_keyboard(int key, bool press)
{
	switch(key) {
	case KB_ESC:
		deactivate_screen();
		break;

	default:
		break;
	}
}

void Renderer::message(MsgAtom msg, ...)
{
	va_list ap;
	va_start(ap, msg);

	for(int i=0; i<NUM_MESSAGES; i++) {
		if(msg == rend->msg[i]) {
			switch(i) {
			case MSG_START:
				rend->cur_x = rend->cur_y = 0;
				rend->in_progress = true;
				memset(rend->pixels, 0, fb_width * fb_height * 4);

				printlog("starting a new rendering!\n");
				break;

			case MSG_GAMMA:
				rend->gamma = va_arg(ap, double);
				rend->inv_gamma = 1.0 / rend->gamma;
				break;

			case MSG_USE_GAMMA:
				rend->use_gamma = va_arg(ap, int);
				break;
			}
		}
	}

	va_end(ap);
}
