#include <string.h>
#include <float.h>
#include "scene.h"
#include "min3d.h"

Scene::Scene()
{
	name = 0;
	active_cam = 0;

	fog_exp = 1.0;

	set_ambient(Vector3(0.05, 0.05, 0.05));

	set_background(Vector3(0.4, 0.3, 0.05), SCN_BG_LOW);
	set_background(Vector3(0.6, 0.6, 0.6), SCN_BG_MID);
	set_background(Vector3(0.1, 0.5, 1.0), SCN_BG_HIGH);
}

Scene::~Scene()
{
	clear();
}

void Scene::clear()
{
	delete [] name;

	size_t i;
	for(i=0; i<objects.size(); i++) {
		delete objects[i];
	}
	for(i=0; i<lights.size(); i++) {
		delete lights[i];
	}
	for(i=0; i<cameras.size(); i++) {
		delete cameras[i];
	}
}

void Scene::set_name(const char *name)
{
	delete [] this->name;
	this->name = new char[strlen(name) + 1];
	strcpy(this->name, name);
}

const char *Scene::get_name() const
{
	return name ? name : "<unknown>";
}

void Scene::set_background(const Vector3 &col, int idx)
{
	if(idx == -1) {
		bg[0] = bg[1] = bg[2] = col;
	} else {
		bg[idx] = col;
	}
}

const Vector3 &Scene::get_background(int idx) const
{
	return bg[idx == -1 ? 0 : idx];
}

#define LERP(a, b, t)	((a) + ((b) - (a)) * t)

Vector3 Scene::get_background(const Ray &ray) const
{
	Vector3 dir = normalize(ray.dir);

	float v;
	if(dir.z == 0.0) {
		v = dir.y > 0.0 ? 1.0 : -1.0;
	} else {
		float angle = atan2(dir.z, dir.y);
		v = angle / M_PI;
	}

	float r = v >= 0.0 ? LERP(bg[1].x, bg[2].x, v) : LERP(bg[1].x, bg[0].x, -v);
	float g = v >= 0.0 ? LERP(bg[1].y, bg[2].y, v) : LERP(bg[1].y, bg[0].y, -v);
	float b = v >= 0.0 ? LERP(bg[1].z, bg[2].z, v) : LERP(bg[1].z, bg[0].z, -v);
	return Vector3(r, g, b);
}

void Scene::set_ambient(const Vector3 &col)
{
	ambient = col;
}

const Vector3 &Scene::get_ambient() const
{
	return ambient;
}


void Scene::add(SceneNode *node)
{
	nodes.push_back(node);

	switch(node->get_type()) {
	case NODE_OBJECT:
		objects.push_back((Object*)node);
		break;

	case NODE_LIGHT:
		lights.push_back((Light*)node);
		break;

	case NODE_CAMERA:
		cameras.push_back((Camera*)node);
		break;

	default:
		break;
	}

	sel.clear();
}

int Scene::get_node_count() const
{
	return (int)nodes.size();
}

int Scene::get_object_count() const
{
	return (int)objects.size();
}

int Scene::get_light_count() const
{
	return (int)lights.size();
}

int Scene::get_camera_count() const
{
	return (int)cameras.size();
}

SceneNode *Scene::get_node(int idx)
{
	return nodes[idx];
}

const SceneNode *Scene::get_node(int idx) const
{
	return nodes[idx];
}

Object *Scene::get_object(int idx)
{
	return objects[idx];
}

const Object *Scene::get_object(int idx) const
{
	return objects[idx];
}

Light *Scene::get_light(int idx)
{
	return lights[idx];
}

const Light *Scene::get_light(int idx) const
{
	return lights[idx];
}

Camera *Scene::get_camera(int idx)
{
	return cameras[idx];
}

const Camera *Scene::get_camera(int idx) const
{
	return cameras[idx];
}

void Scene::set_active_camera(Camera *cam)
{
	active_cam = cam;
}

Camera *Scene::get_active_camera() const
{
	return active_cam;
}

void Scene::draw() const
{
	if(active_cam) {
		// TODO
	}

	int num_nodes = get_node_count();
	for(int i=0; i<num_nodes; i++) {
		bool selected = false;

		for(size_t j=0; j<sel.size(); j++) {
			if(sel[j] == i) {
				selected = true;
				break;
			}
		}

		nodes[i]->draw(selected);
	}
}

void Scene::select(int s)
{
	for(size_t i=0; i<sel.size(); i++) {
		if(sel[i] == s) {
			return;
		}
	}
	sel.push_back(s);
}

void Scene::clear_selection()
{
	sel.clear();
}

int Scene::get_selection_count() const
{
	return (int)sel.size();
}

int Scene::get_selection(int idx) const
{
	return idx >= 0 && idx < (int)sel.size() ? sel[idx] : -1;
}

bool Scene::intersect(const Ray &ray, RayHit *hit) const
{
	if(hit) {
		hit->dist = FLT_MAX;
		hit->obj = 0;
	}

	int num_obj = (int)objects.size();
	for(int i=0; i<num_obj; i++) {
		RayHit tmphit;
		if(objects[i]->intersect(ray, hit ? &tmphit : 0)) {
			if(!hit) return true;
			if(tmphit.dist < hit->dist) {
				*hit = tmphit;
			}
		}
	}
	return hit && hit->obj;
}

bool Scene::load(const char *fname)
{
	FILE *fp = fopen(fname, "r");
	if(!fp) {
		return false;
	}
	bool res = load(fp);
	fclose(fp);
	return res;
}

bool Scene::save(const char *fname) const
{
	FILE *fp = fopen(fname, "w");
	if(!fp) {
		return false;
	}
	bool res = save(fp);
	fclose(fp);
	return res;
}

bool Scene::load(FILE *fp)
{
	return false;	// TODO
}

bool Scene::save(FILE *fp) const
{
	return false;	// TODO
}
