#include <string.h>
#include <assert.h>
#include "snode.h"
#include "min3d.h"


SceneNode::SceneNode()
{
	parent = 0;
	name = 0;
	type = NODE_NULL;
	scale = Vector3(1, 1, 1);
	invalidate();
}

SceneNode::~SceneNode()
{
	delete [] name;
}

void SceneNode::set_name(const char *name)
{
	this->name = new char[strlen(name) + 1];
	strcpy(this->name, name);
}

const char *SceneNode::get_name() const
{
	return name ? name : "<unnamed>";
}

NodeType SceneNode::get_type() const
{
	return type;
}

SceneNode *SceneNode::get_parent()
{
	return parent;
}

const SceneNode *SceneNode::get_parent() const
{
	return parent;
}

void SceneNode::add_child(SceneNode *child)
{
	children.push_back(child);
	child->parent = this;
	invalidate();
}

void SceneNode::remove_child(SceneNode *child)
{
	// TODO
	invalidate();
}

int SceneNode::get_children_count() const
{
	return (int)children.size();
}

SceneNode *SceneNode::get_child(int idx)
{
	if(idx >= 0 && idx < get_children_count()) {
		return children[idx];
	}
	return 0;
}

const SceneNode *SceneNode::get_child(int idx) const
{
	if(idx >= 0 && idx < get_children_count()) {
		return children[idx];
	}
	return 0;
}



void SceneNode::set_position(const Vector3 &pos)
{
	this->pos = pos;
	invalidate();
}

Vector3 SceneNode::get_node_position() const
{
	return pos;
}

void SceneNode::set_rotation(const Quat &quat)
{
	rot = quat;
	invalidate();
}

Quat SceneNode::get_node_rotation() const
{
	return rot;
}

void SceneNode::set_scaling(const Vector3 &scale)
{
	this->scale = scale;
	invalidate();
}

Vector3 SceneNode::get_node_scaling() const
{
	return scale;
}

// these take hierarchy into account
Vector3 SceneNode::get_position() const
{
	return transform(get_matrix(), Vector3(0, 0, 0));
}

Quat SceneNode::get_rotation() const
{
	if(parent) {
		return parent->get_rotation() * rot;
	}
	return rot;
}

Vector3 SceneNode::get_scaling() const
{
	if(parent) {
		return parent->get_scaling() * scale;
	}
	return scale;
}

void SceneNode::set_pivot(const Vector3 &pivot)
{
	this->pivot = pivot;
	invalidate();
}

Vector3 SceneNode::get_pivot() const
{
	return pivot;
}

const Matrix4x4 &SceneNode::get_matrix() const
{
	calc_matrix();
	return xform;
}

const Matrix4x4 &SceneNode::get_inv_matrix() const
{
	calc_inv_matrix();
	return inv_xform;
}

void SceneNode::invalidate() const
{
	xform_valid = inv_xform_valid = false;
}

// TODO: hierarchy
void SceneNode::calc_matrix() const
{
	if(xform_valid) return;

	xform.set_identity();
	xform.translate(pivot.x, pivot.y, pivot.z);
	xform = xform * rot.get_matrix();
	xform.translate(pos.x, pos.y, pos.z);
	xform.scale(scale.x, scale.y, scale.z);
	xform.translate(-pivot.x, -pivot.y, -pivot.z);

	xform_valid = true;
}

void SceneNode::calc_inv_matrix() const
{
	if(inv_xform_valid) return;

	calc_matrix();

	inv_xform = xform.inverse();
	inv_xform_valid = true;
}

void SceneNode::pre_draw() const
{
	m3d_matrix_mode(M3D_MODELVIEW);
	m3d_push_matrix();
	m3d_mult_transpose_matrix(get_matrix()[0]);
}

void SceneNode::post_draw() const
{
	m3d_matrix_mode(M3D_MODELVIEW);
	m3d_pop_matrix();
}

void SceneNode::draw(bool emph) const
{
	if(emph) {
		float avg_scale = (scale.x + scale.y + scale.z) / 3.0;
		m3d_push_matrix();
		m3d_scale(avg_scale / scale.x, avg_scale / scale.y, avg_scale / scale.z);

		m3d_begin(M3D_LINES);
		m3d_color(1, 0, 0);
		m3d_vertex(0, 0, 0);
		m3d_vertex(0.5, 0, 0);
		m3d_color(0, 1, 0);
		m3d_vertex(0, 0, 0);
		m3d_vertex(0, 0.5, 0);
		m3d_color(0, 0, 1);
		m3d_vertex(0, 0, 0);
		m3d_vertex(0, 0, 0.5);
		m3d_end();

		m3d_pop_matrix();

		m3d_color(1, 0.8, 0.1);
	} else {
		m3d_color(0.9, 0.9, 0.9);
	}
}

bool SceneNode::intersect(const Ray &ray, RayHit *hit) const
{
	return false;
}
