#ifndef SCENE_NODE_H_
#define SCENE_NODE_H_

#include "vector.h"
#include "vmath.h"
#include "quat.h"
#include "vmathmat.h"
#include "vmathray.h"
#include "raytrace.h"

enum NodeType {
	NODE_NULL,
	NODE_OBJECT,
	NODE_LIGHT,
	NODE_CAMERA
};

class SceneNode {
protected:
	char *name;
	NodeType type;

	Vector3 pos;
	Quat rot;
	Vector3 scale;
	Vector3 pivot;

	mutable Matrix4x4 xform, inv_xform;
	mutable bool xform_valid, inv_xform_valid;

	vector<SceneNode*> children;
	SceneNode *parent;

	SceneNode(const SceneNode &node) {}
	SceneNode &operator =(const SceneNode &node) { return *this; }

	void invalidate() const;
	virtual void calc_matrix() const;
	virtual void calc_inv_matrix() const;

	virtual void pre_draw() const;
	virtual void post_draw() const;

public:
	SceneNode();
	virtual ~SceneNode();

	virtual void set_name(const char *name);
	virtual const char *get_name() const;

	virtual NodeType get_type() const;

	virtual SceneNode *get_parent();
	virtual const SceneNode *get_parent() const;

	// children management
	virtual void add_child(SceneNode *child);
	virtual void remove_child(SceneNode *child);

	virtual int get_children_count() const;
	virtual SceneNode *get_child(int idx);
	virtual const SceneNode *get_child(int idx) const;

	virtual void set_position(const Vector3 &pos);
	virtual Vector3 get_node_position() const;

	virtual void set_rotation(const Quat &quat);
	virtual Quat get_node_rotation() const;

	virtual void set_scaling(const Vector3 &scale);
	virtual Vector3 get_node_scaling() const;

	// these take hierarchy into account
	virtual Vector3 get_position() const;
	virtual Quat get_rotation() const;
	virtual Vector3 get_scaling() const;

	virtual void set_pivot(const Vector3 &pivot);
	virtual Vector3 get_pivot() const;

	virtual const Matrix4x4 &get_matrix() const;
	virtual const Matrix4x4 &get_inv_matrix() const;

	virtual void draw(bool emph = false) const;

	virtual bool intersect(const Ray &ray, RayHit *hit = 0) const;
};

#endif	/* SCENE_NODE_H_ */
