#ifndef VMATH_H_
#define VMATH_H_

#include <math.h>
#include "vmathmat.h"

#define DEG2RAD(x)	(M_PI * (x) / 180.0)
#define RAD2DEG(x)	(180.0 * (x) / M_PI)

class Vector3 {
public:
	float x, y, z;

	Vector3() : x(0), y(0), z(0) {}
	Vector3(float xx, float yy, float zz) : x(xx), y(yy), z(zz) {}

	float length_sq() const { return x * x + y * y + z * z; }
	float length() const { return sqrt(x * x + y * y + z * z); }

	void normalize()
	{
		float len = length();
		if(len != 0.0) {
			x /= len;
			y /= len;
			z /= len;
		}
	}

	float &operator [](int idx) { return idx == 2 ? z : (idx == 1 ? y : x); }
	const float &operator [](int idx) const { return idx == 2 ? z : (idx == 1 ? y : x); }
};

inline Vector3 normalize(const Vector3 &v)
{
	float len = v.length();
	if(len != 0.0) {
		return Vector3(v.x / len, v.y / len, v.z / len);
	}
	return v;
}

inline Vector3 operator +(const Vector3 &a, const Vector3 &b)
{
	return Vector3(a.x + b.x, a.y + b.y, a.z + b.z);
}

inline Vector3 operator -(const Vector3 &a, const Vector3 &b)
{
	return Vector3(a.x - b.x, a.y - b.y, a.z - b.z);
}

inline Vector3 operator -(const Vector3 &v)
{
	return Vector3(-v.x, -v.y, -v.z);
}

inline Vector3 operator *(const Vector3 &a, const Vector3 &b)
{
	return Vector3(a.x * b.x, a.y * b.y, a.z * b.z);
}

inline Vector3 operator *(const Vector3 &v, float s)
{
	return Vector3(v.x * s, v.y * s, v.z * s);
}

inline Vector3 operator /(const Vector3 &v, float s)
{
	return Vector3(v.x / s, v.y / s, v.z / s);
}

inline float dot(const Vector3 &a, const Vector3 &b)
{
	return a.x * b.x + a.y * b.y + a.z * b.z;
}

inline Vector3 cross(const Vector3 &a, const Vector3 &b)
{
	return Vector3(a.y * b.z - a.z * b.y,
			a.z * b.x - a.x * b.z,
			a.x * b.y - a.y * b.x);
}

inline Vector3 transform(const Matrix4x4 &m, const Vector3 &v)
{
	float x = m.m[0][0] * v.x + m.m[0][1] * v.y + m.m[0][2] * v.z + m.m[0][3];
	float y = m.m[1][0] * v.x + m.m[1][1] * v.y + m.m[1][2] * v.z + m.m[1][3];
	float z = m.m[2][0] * v.x + m.m[2][1] * v.y + m.m[2][2] * v.z + m.m[2][3];
	return Vector3(x, y, z);
}

inline Vector3 lerp(const Vector3 &a, const Vector3 &b, float t)
{
	return Vector3(a.x + (b.x - a.x) * t,
			a.y + (b.y - a.y) * t,
			a.z + (b.z - a.z) * t);
}

inline Vector3 reflect(const Vector3 &v, const Vector3 &n)
{
	float vdotn = dot(v, n);
	return n * vdotn * 2.0 - v;
}

inline Vector3 refract(const Vector3 &v, const Vector3 &n, float ior)
{
	float cos_inc = dot(v, -n);
	float radical = 1.0 + ior * ior * (cos_inc * cos_inc - 1.0);

	if(radical < 0.0) {	// total internal reflection
		return -reflect(v, n);
	}

	float beta = ior * cos_inc - sqrt(radical);
	return v * ior + n * beta;
}

// ---- Vector4 ----

class Vector4 {
public:
	float x, y, z, w;

	Vector4() : x(0), y(0), z(0), w(1.0) {}
	Vector4(const Vector3 &v) : x(v.x), y(v.y), z(v.z), w(1.0) {}
	Vector4(float xx, float yy, float zz, float ww) : x(xx), y(yy), z(zz), w(ww) {}

	float length_sq() const { return x * x + y * y + z * z + w * w; }
	float length() const { return sqrt(x * x + y * y + z * z + w * w); }

	void normalize()
	{
		float len = length();
		if(len != 0.0) {
			x /= len;
			y /= len;
			z /= len;
			w /= len;
		}
	}

	float &operator [](int idx)
	{
		return idx == 3 ? w : (idx == 2 ? z : (idx == 1 ? y : x));
	}
	const float &operator [](int idx) const
	{
		return idx == 3 ? w : (idx == 2 ? z : (idx == 1 ? y : x));
	}
};

inline Vector4 operator +(const Vector4 &a, const Vector4 &b)
{
	return Vector4(a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w);
}

inline Vector4 operator -(const Vector4 &a, const Vector4 &b)
{
	return Vector4(a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w);
}

inline Vector4 operator -(const Vector4 &v)
{
	return Vector4(-v.x, -v.y, -v.z, -v.w);
}

inline Vector4 operator *(const Vector4 &v, float s)
{
	return Vector4(v.x * s, v.y * s, v.z * s, v.w * s);
}

inline Vector4 operator /(const Vector4 &v, float s)
{
	return Vector4(v.x / s, v.y / s, v.z / s, v.w / s);
}

inline float dot(const Vector4 &a, const Vector4 &b)
{
	return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;
}

inline Vector4 transform(const Matrix4x4 &m, const Vector4 &v)
{
	float x = m.m[0][0] * v.x + m.m[0][1] * v.y + m.m[0][2] * v.z + m.m[0][3] * v.w;
	float y = m.m[1][0] * v.x + m.m[1][1] * v.y + m.m[1][2] * v.z + m.m[1][3] * v.w;
	float z = m.m[2][0] * v.x + m.m[2][1] * v.y + m.m[2][2] * v.z + m.m[2][3] * v.w;
	float w = m.m[3][0] * v.x + m.m[3][1] * v.y + m.m[3][2] * v.z + m.m[3][3] * v.w;
	return Vector4(x, y, z, w);
}

inline Vector4 lerp(const Vector4 &a, const Vector4 &b, float t)
{
	return Vector4(a.x + (b.x - a.x) * t,
			a.y + (b.y - a.y) * t,
			a.z + (b.z - a.z) * t,
			a.w + (b.w - a.w) * t);
}


#endif	// VMATH_H_
