#ifndef VMATH_MATRIX_H_
#define VMATH_MATRIX_H_

#include <math.h>

#ifndef M_PI
#define M_PI	3.141592653
#endif

class Vector3;

class Matrix4x4 {
public:
	float m[4][4];

	Matrix4x4()
	{
		set_identity();
	}

	Matrix4x4(float m00, float m01, float m02, float m03,
			float m10, float m11, float m12, float m13,
			float m20, float m21, float m22, float m23,
			float m30, float m31, float m32, float m33)
	{
		m[0][0] = m00; m[0][1] = m01; m[0][2] = m02; m[0][3] = m03;
		m[1][0] = m10; m[1][1] = m11; m[1][2] = m12; m[1][3] = m13;
		m[2][0] = m20; m[2][1] = m21; m[2][2] = m22; m[2][3] = m23;
		m[3][0] = m30; m[3][1] = m31; m[3][2] = m32; m[3][3] = m33;
	}

	inline void set_identity();
	inline void translate(float x, float y, float z);
	inline void rotate(float angle, float x, float y, float z);
	inline void scale(float x, float y, float z);
	inline void perspective(float vfov, float aspect, float znear, float zfar);
	void lookat(const Vector3 &pos, const Vector3 &targ, const Vector3 &up);

	float *operator [](int idx) { return m[idx]; }
	const float *operator [](int idx) const { return m[idx]; }

	void transpose();

	float determinant() const;
	Matrix4x4 adjoint() const;
	Matrix4x4 inverse() const;
};

inline Matrix4x4 operator *(const Matrix4x4 &a, const Matrix4x4 &b)
{
	Matrix4x4 res;
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			res[i][j] = a[i][0] * b[0][j] + a[i][1] * b[1][j] +
				a[i][2] * b[2][j] + a[i][3] * b[3][j];
		}
	}
	return res;
}

inline Matrix4x4 operator *(const Matrix4x4 &mat, float scalar)
{
	Matrix4x4 res;

	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			res.m[i][j] = mat.m[i][j] * scalar;
		}
	}
	return res;
}

inline void Matrix4x4::set_identity()
{
	m[0][0] = m[1][1] = m[2][2] = m[3][3] = 1.0;
	m[0][1] = m[0][2] = m[0][3] = m[1][2] = m[1][3] = m[2][3] = 0.0;
	m[1][0] = m[2][0] = m[3][0] = m[2][1] = m[3][1] = m[3][2] = 0.0;
}

inline void Matrix4x4::translate(float x, float y, float z)
{
	Matrix4x4 m(1, 0, 0, x, 0, 1, 0, y, 0, 0, 1, z, 0, 0, 0, 1);
	*this = *this * m;
}

inline void Matrix4x4::rotate(float angle, float x, float y, float z)
{
	float sina = (float)sin(angle);
	float cosa = (float)cos(angle);
	float rcosa = 1.0f - cosa;
	float nxsq = x * x;
	float nysq = y * y;
	float nzsq = z * z;

	Matrix4x4 m;
	m[0][0] = nxsq + (1.0f - nxsq) * cosa;
	m[0][1] = x * y * rcosa - z * sina;
	m[0][2] = x * z * rcosa + y * sina;

	m[1][0] = x * y * rcosa + z * sina;
	m[1][1] = nysq + (1.0f - nysq) * cosa;
	m[1][2] = y * z * rcosa - x * sina;

	m[2][0] = x * z * rcosa - y * sina;
	m[2][1] = y * z * rcosa + x * sina;
	m[2][2] = nzsq + (1.0f - nzsq) * cosa;

	*this = *this * m;
}

inline void Matrix4x4::scale(float x, float y, float z)
{
	Matrix4x4 m(x, 0, 0, 0, 0, y, 0, 0, 0, 0, z, 0, 0, 0, 0, 1);
	*this = *this * m;
}

inline void Matrix4x4::perspective(float vfov, float aspect, float znear, float zfar)
{
	float f = 1.0f / tan(vfov * 0.5f);
	float dz = znear - zfar;

	Matrix4x4 m;
	m[0][0] = f / aspect;
	m[1][1] = f;
	m[2][2] = (zfar + znear) / dz;
	m[3][2] = -1.0f;
	m[2][3] = 2.0f * zfar * znear / dz;
	m[3][3] = 0.0f;

	*this = *this * m;
}

inline Matrix4x4 normal_matrix(const Matrix4x4 &m)
{
	return Matrix4x4(m[0][0], m[0][1], m[0][2], 0,
					m[1][0], m[1][1], m[1][2], 0,
					m[2][0], m[2][1], m[2][2], 0,
					m[3][0], m[3][1], m[3][2], 1);
}

#endif	// VMATH_MATRIX_H_
